/*
	Engineering Test: Boggle Solver
	Gordon Lee

*******************************************************************************

	@filename	BoggleSolver.cpp

	@description
		Function for solving all valid words in a Boggle board from a dictionary

	@author		Gordon Lee
*/

// Includes
#include "BoggleSolver.h"

// Namespace
namespace Test
{
////////////////////////////////////////////////////////////////////////////////
/**
	Boggle Solver

	@param cBoard *_board
	@param cDictionary *_dictionary

	@return cList<std::string> *
*/
cList<std::string> *BoggleSolver(cBoard *_board, cDictionary *_dictionary)
{
	cList<std::string> *_validWords = new cList<std::string>;
	char *_wordBuffer = new char[_board->Size() + 1];
	char *_boardState = new char[_board->Size() + 1];
	memset(_wordBuffer, '\0', _board->Size() + 1);
	memset(_boardState, 'O', _board->Size() + 1);
	_boardState[_board->Size()] = '\0';

	sBoggleSolverInfo _info;
	_info.m_board = _board;
	_info.m_boardState = _boardState;
	_info.m_currBoardPosR = -1;
	_info.m_currBoardPosC = -1;
	_info.m_dictionary = _dictionary;
	_info.m_validWords = _validWords;
	_info.m_wordBuffer = _wordBuffer;
	_info.m_wordIndex = 0;

	RecursiveBoggleSolver(_info, _dictionary->Trie()->Head());

	delete [] _wordBuffer;
	delete [] _boardState;

	return _validWords;
}

/**
	Recursive Boggle Solver

	@param sBoggleSolverInfo &_info
	@param sTrieNode *_nodeRef
*/
void RecursiveBoggleSolver(sBoggleSolverInfo &_info, sTrieNode *_nodeRef)
{
	sTrieNode *_temp = _nodeRef->m_child;
	int _r = 0;
	int _c = 0;
	int _wordIndex = 0;
	bool _bAlreadyFound = false;

	// Exit condition
	if (_temp == NULL)
		return;

	// Search through siblings
	for (; _temp != NULL; _temp = _temp->m_sibling)
	{
		// If first letter, search entire board
		if (_info.m_wordIndex == 0)
		{
			for (_r = 0; _r < _info.m_board->Rows(); ++_r)
			{
				for (_c = 0; _c < _info.m_board->Cols(); ++_c)
				{
					// If letter found on board
					if (_temp->m_char == _info.m_board->GetChar(_r, _c))
					{
						_wordIndex = _info.m_wordIndex;
						_info.m_boardState[(_r * _info.m_board->Cols()) + _c] = 'S';
						_info.m_currBoardPosR = _r;
						_info.m_currBoardPosC = _c;
						_info.m_wordBuffer[_info.m_wordIndex++] = _temp->m_char;

						// Go to next letter
						RecursiveBoggleSolver(_info, _temp);

						// Undo
						_info.m_boardState[(_r * _info.m_board->Cols()) + _c] = 'O';
						_info.m_currBoardPosR = _r;
						_info.m_currBoardPosC = _c;
						_info.m_wordBuffer[_wordIndex] = '\0';
						_info.m_wordIndex = _wordIndex;
					}
				}
			}
		}
		// If subsequent letter, only search neighbors and do not reuse letters
		else
		{
			_bAlreadyFound = false;
			_r = _info.m_currBoardPosR;
			_c = _info.m_currBoardPosC;

			// Iterate through all neighbors
			for (int i = _r - 1; i <= _r + 1; ++i)
			{
				for (int j = _c - 1; j <= _c + 1; ++j)
				{
					if (!_bAlreadyFound &&
						i >= 0 && j >= 0 && i <= _info.m_board->Rows() - 1 && j <= _info.m_board->Cols() - 1 &&
						_temp->m_char == _info.m_board->GetChar(i, j) &&
						_info.m_boardState[(i * _info.m_board->Cols()) + j] == 'O')
					{
						_wordIndex = _info.m_wordIndex;
						_info.m_boardState[(i * _info.m_board->Cols()) + j] = 'X';
						_info.m_currBoardPosR = i;
						_info.m_currBoardPosC = j;
						_info.m_wordBuffer[_info.m_wordIndex++] = _temp->m_char;

						// If end of word
						if (_temp->m_bEOW)
						{
							_info.m_wordBuffer[_info.m_wordIndex] = '\0';
							if (strlen(_info.m_wordBuffer) >= _info.m_dictionary->MinLetters() && !_info.m_validWords->Find(_info.m_wordBuffer))
								_info.m_validWords->Push_Back(_info.m_wordBuffer);
							_bAlreadyFound = true;
						}

						// Go to next letter
						RecursiveBoggleSolver(_info, _temp);

						// Undo
						_info.m_boardState[(i * _info.m_board->Cols()) + j] = 'O';
						_info.m_currBoardPosR = _r;
						_info.m_currBoardPosC = _c;
						_info.m_wordBuffer[_wordIndex] = '\0';
						_info.m_wordIndex = _wordIndex;
					}
				}
			}
		}
	}
}
////////////////////////////////////////////////////////////////////////////////
}
