/*
	Engineering Test: Boggle Solver
	Gordon Lee

*******************************************************************************

	@filename	cMemoryManager.cpp

	@description
		Class for a memory manager for pooling and handling memory.

	@author		Gordon Lee
*/

// Includes
#include "cMemoryManager.h"

namespace Test
{
////////////////////////////////////////////////////////////////////////////////
/**
	Constructor
*/
cMemoryManager::cMemoryManager()
: m_numAllocations(0), m_numFrees(0), m_numPools(0),
	m_numUsedMemory(0), m_maxMemoryUsed(0), m_totalMemoryAllocated(0)
{
}

/**
	Deconstructor
*/
cMemoryManager::~cMemoryManager()
{
	MemoryPoolMap::iterator _iter = m_memoryPoolMap.begin();
	MemoryPoolMap::iterator _iend = m_memoryPoolMap.end();
	for (; _iter != _iend; ++_iter)
	{
		m_totalMemoryAllocated += (_iter->second)->GetTotalMemoryAllocated();
		delete _iter->second;
	}
	m_memoryPoolMap.clear();
}

/**
	Instance

	@return cMemoryManager *
*/
cMemoryManager *cMemoryManager::Instance(void)
{
	static cMemoryManager *_memoryManager = new cMemoryManager();
	return _memoryManager;
}

/**
	Create

	@return cMemoryManager *
*/
cMemoryManager *cMemoryManager::Create(void)
{
	return Instance();
}

/**
	Destroy
*/
void cMemoryManager::Destroy(void)
{
	delete Instance();
}

/**
	Allocate

	@param size_t _objectSize

	@return void *
*/
void *cMemoryManager::Allocate(size_t _objectSize)
{
	MemoryPoolMap::iterator _iter = m_memoryPoolMap.find(_objectSize);

	if (_iter == m_memoryPoolMap.end())
	{
		_iter = CreateMemoryPool(_objectSize);
	}

	++m_numAllocations;
	m_numUsedMemory += (unsigned)_objectSize;
	if (m_maxMemoryUsed < m_numUsedMemory)
	{
		m_maxMemoryUsed = m_numUsedMemory;
	}

	return (_iter->second)->Allocate();
}

/**
	Free

	@param size_t _objectSize
	@param void *_object
*/
void cMemoryManager::Free(size_t _objectSize, void *_object)
{
	MemoryPoolMap::iterator _iter = m_memoryPoolMap.find(_objectSize);

	if (_iter != m_memoryPoolMap.end())
	{
		(_iter->second)->Free(_object);

		++m_numFrees;
		m_numUsedMemory -= (unsigned)_objectSize;
	}
}

/**
	Create Memory Pool

	@param size_t _objectSize

	@return MemoryPoolMap::iterator
*/
MemoryPoolMap::iterator cMemoryManager::CreateMemoryPool(size_t _objectSize)
{
	cMemoryPool *_pool = new cMemoryPool((unsigned)_objectSize);

	std::pair<MemoryPoolMap::iterator, bool> ret;
	ret = m_memoryPoolMap.insert(MemoryPoolMap::value_type(_objectSize, _pool));

	++m_numPools;

	return ret.first;
}
////////////////////////////////////////////////////////////////////////////////
}
